import torch

from torch_hep.lorentz import LorentzTensor, MomentumTensor


def test_LorentzTensor():
    short = LorentzTensor([111549, 35202.7, -46507.4, 94552.1])
    long  = LorentzTensor([[111549, 35202.7, -46507.4, 94552.1],
                                        [86549.2, 12443.8, 81453.1, 25407.5],
                                        [86799.1, 12423.2, 81499.2, 25411.3]])
    
    assert torch.all(torch.eq((short*2).values,(short+short).values))
    assert torch.all(torch.eq((long*2).values,(long+long).values))
    assert torch.all(torch.isclose((short-short/2).values,(short/2).values,1.))
    assert torch.all(torch.isclose((long-long/2).values,(long/2).values,1.))


def test_MomentumTensor():
    short = MomentumTensor([111549, 35202.7, -46507.4, 94552.1])
    long  = MomentumTensor([[111549, 35202.7, -46507.4, 94552.1],
                            [86549.2, 12443.8, 81453.1, 25407.5],
                            [86799.1, 12423.2, 81499.2, 25411.3]])
    short_EEtaPhiPt = MomentumTensor.EEtaPhiPt([111549.0, 1.2600811625519022, -0.9228767507929038, 58328.10936461082])
    long_EEtaPhiPt  = MomentumTensor.EEtaPhiPt([[111549.0, 1.2600811625519022, -0.9228767507929038, 58328.10936461082],
                                                [86549.2, 0.3036619846021553, 1.4191959214152075, 82398.15324417168],
                                                [86799.1, 0.30355425211172465, 1.4195273814629104, 82440.61801612115]])
    short_MEtaPhiPt = MomentumTensor.MEtaPhiPt([10045.468856155943, 1.2600811625519022, -0.9228767507929038, 58328.10936461082])
    long_MEtaPhiPt  = MomentumTensor.MEtaPhiPt([[10045.468856155943, 1.2600811625519022, -0.9228767507929038, 58328.10936461082],
                                                [7467.751089852902, 0.3036619846021553, 1.4191959214152075, 82398.15324417168],
                                                [9586.140737544087, 0.30355425211172465, 1.4195273814629104, 82440.61801612115]])

    assert torch.all(torch.isclose(short[0],short_EEtaPhiPt.e))
    assert torch.all(torch.isclose(long[0],long_EEtaPhiPt.e))
    assert torch.all(torch.isclose(short[1],short_EEtaPhiPt.px))
    assert torch.all(torch.isclose(long[1],long_EEtaPhiPt.px))
    assert torch.all(torch.isclose(short[2],short_EEtaPhiPt.py))
    assert torch.all(torch.isclose(long[2],long_EEtaPhiPt.py))
    assert torch.all(torch.isclose(short[3],short_EEtaPhiPt.pz))
    assert torch.all(torch.isclose(long[3],long_EEtaPhiPt.pz))

    assert torch.all(torch.isclose(short[0],short_MEtaPhiPt.e))
    assert torch.all(torch.isclose(long[0],long_MEtaPhiPt.e))
    assert torch.all(torch.isclose(short[1],short_MEtaPhiPt.px))
    assert torch.all(torch.isclose(long[1],long_MEtaPhiPt.px))
    assert torch.all(torch.isclose(short[2],short_MEtaPhiPt.py))
    assert torch.all(torch.isclose(long[2],long_MEtaPhiPt.py))
    assert torch.all(torch.isclose(short[3],short_MEtaPhiPt.pz))
    assert torch.all(torch.isclose(long[3],long_MEtaPhiPt.pz))

    assert torch.isclose(short.p,torch.tensor([111095.76029921215]))
    assert torch.all(torch.isclose(long.p,torch.tensor([[111095.76029921215],[86226.42700645783],[86268.12659708102]])))

    assert torch.isclose(short.pt,torch.tensor([58328.10936461082]))
    assert torch.all(torch.isclose(long.pt,torch.tensor([[58328.10936461082],[82398.15324417168],[82440.61801612115]])))

    assert torch.isclose(short.m,torch.tensor([10045.468856155943]))
    assert torch.all(torch.isclose(long.m,torch.tensor([[10045.468856155943],[7467.751089852902],[9586.140737544087]])))



import torch

from torch import Tensor


class GraphBuilder():
    r""":class:`GraphBuilder`
    """
    _graph = dict[Tensor]
    def __init__(self, device='cpu'):
        self._graph = self._graph()
        self.values = self._graph
        self.device = device

    def __getitem__(self, key):
        return self._graph[key]

    def __setitem__(self, key, value):
        self._graph[key] = value

    def add_asGlobal(self, key: str, **attr):
        if key not in self._graph:
            self._graph.update({key: torch.tensor([*attr.values()],device=self.device)})
            self._graph.update({key+'_features': [*attr.keys()]})
        elif key in self._graph:
            self[key] = torch.cat((self[key],torch.tensor([*attr.values()],device=self.device)),dim=0)
            self[key+'_features'] += [*attr.keys()]

    def add_asNode(self, key: str, **attr):
        if key not in self._graph:
            self._graph.update({key: torch.tensor([*attr.values()],device=self.device).transpose(1,0)})
            self._graph.update({key+'_features': [*attr.keys()]})
        elif key in self._graph:
            self[key] = torch.cat((self[key],torch.tensor([*attr.values()],device=self.device).transpose(1,0)),dim=1)
            self[key+'_features'] += [*attr.keys()]

    def add_asEdge(self, key: str, **attr):
        if 'edge_index' not in self._graph:
            if 'index' not in attr:
                raise ValueError("'index' must be included")
            else:
                self._graph.update({'edge_index': torch.tensor(attr.get('index'),device=self.device).transpose(0,1)})
                del attr['index']
        if key not in self._graph:
            self._graph.update({key: torch.tensor([*attr.values()],device=self.device).transpose(1,0)})
            self._graph.update({key+'_features': [*attr.keys()]})
        elif key in self._graph:
            self[key] = torch.cat((self[key],torch.tensor([*attr.values()],device=self.device).transpose(1,0)),dim=1)
            self[key+'_features'] += [*attr.keys()]




import torch

from torch import Tensor


class LorentzTensor(object):
    r""":class:`LorentzTensor`

    Args:
        input (torch.tensor): input tensor with size `torch.Size([N,4])`.
        device (str): device is used to store tensors (default 'cpu').
    """
    def __init__(self, input, **kwargs):
        if torch.is_tensor(input) == True:
            self.values = input.clone()
        else:
            self.values = torch.tensor(input,**kwargs)
        self.device = self.values.device
        self._minkowski = torch.tensor([1,-1,-1,-1],device=self.device)
        if self.values.ndim == 1:
            if self.values.size() == torch.Size([4]):
                self.type = 'single'
                self.size = 1
            else:
                raise TypeError("For 1 dimension tensor, expect size 4.")
        else:
            if self.values.size(dim=1) == 4:
                self.type = 'long'
                self.size = self.values.size(dim=0)
            else:
                raise TypeError("Expect size 4 in dimension 1.")

    def __add__(self, other):
        return self.__class__(torch.add(self.values, other.values))

    def __sub__(self, other):
        return self.__class__(torch.sub(self.values, other.values))

    def __mul__(self, other):
        return self.__class__(torch.multiply(other,self.values))
    
    def __rmul__(self, other):
        return self * other
    
    def __truediv__(self, other):
        return self.__class__(torch.div(self.values,other,rounding_mode='floor'))
    
    def __floordiv__(self, other):
        return self.__class__(torch.div(self.values,other,rounding_mode='floor'))
    
    def __div__(self, other):
        return self.__truediv__(other)

    def __getitem__(self, index):
        return self.select(0, index)
    
    def __len__(self):
        return self.size

    def select(self, dim, index):
        if dim == 0:
            return self.values.select(0,index)
        else:
            return self.values.select(dim,index).reshape((-1,1))

    def dot(self, other):
        if (other.values).size() == (self.values).size():
            return torch.sum(self.values * self._minkowski * other.values, dim=1).reshape(-1,1)
        else:
            raise ValueError("two 'LorentzTensor' must have same size in dim=0")

    def to_tensor(self):
        return self.values

    def to_list(self):
        return self.values.tolist()

    @property
    def mag2(self):
        return self.dot(self)

    @property
    def mag(self):
        return torch.sqrt(torch.abs(self.mag2))
    
    @property
    def trans(self):
        return torch.sqrt(self.select(-1,1)**2+self.select(-1,2)**2).reshape(-1,1)

    @property
    def phi(self):
        return torch.atan2(self.select(-1,2),self.select(-1,1)).reshape(-1,1)
        
    @property
    def theta(self):
        return torch.atan2(self.trans,self.select(-1,3)).reshape(-1,1)

    @property
    def eta(self):
        return -torch.log(torch.tan(self.theta/ 2)).reshape(-1,1)


class MomentumTensor(LorentzTensor):
    r""":class:`MomentumTensor`

    Args:
        input (torch.tensor): input tensor with size `torch.Size([N,4])`.
    """
    @staticmethod
    def EEtaPhiPt(input: Tensor):
        Px = (torch.cos(input.select(-1,2))*input.select(-1,3)).reshape([-1,1])
        Py = (torch.sin(input.select(-1,2))*input.select(-1,3)).reshape([-1,1])
        Pz = torch.div(input.select(-1,3),torch.tan(2*torch.arctan(torch.exp(-input.select(-1,1))))).reshape([-1,1])
        return MomentumTensor(torch.cat((input.select(-1,0).reshape([-1,1]),Px,Py,Pz),dim=1))
    
    @staticmethod
    def MEtaPhiPt(input: Tensor):
        Px = (torch.cos(input.select(-1,2))*input.select(-1,3)).reshape([-1,1])
        Py = (torch.sin(input.select(-1,2))*input.select(-1,3)).reshape([-1,1])
        Pz = torch.div(input.select(-1,3),torch.tan(2*torch.arctan(torch.exp(-input.select(-1,1))))).reshape([-1,1])
        E  = torch.sqrt((input.select(-1,0).reshape([-1,1]))**2+Px**2+Py**2+Pz**2)
        return MomentumTensor(torch.cat((E,Px,Py,Pz),dim=1))
    
    @property
    def e(self):
        return self.select(-1,0)

    @property
    def px(self):
        return self.select(-1,1)

    @property
    def py(self):
        return self.select(-1,2)
    
    @property
    def pz(self):
        return self.select(-1,3)
    
    @property
    def p2(self):
        return self.px**2 + self.py**2 + self.pz**2

    @property
    def p(self):
        return torch.sqrt(self.p2)
    
    @property
    def pt(self):
        return self.trans

    @property
    def m2(self):
        return self.mag2
    
    @property
    def m(self):
        return self.mag




from setuptools import find_packages, setup


__version__ = '0.0.3'
URL = 'https://github.com/tzuhanchang/pytorch_hep'

install_requires = [
    'torch',
    'torch_geometric'
]

setup(
    name='torch_hep',
    version=__version__,
    description='Pytorch for high energy physics',
    author='Zihan Zhang',
    author_email='zihan.zhang@cern.ch',
    url=URL,
    download_url=f'{URL}/archive/{__version__}.tar.gz',
    keywords=[
        'pytorch',
    ],
    python_requires='>=3.7',
    install_requires=install_requires,
    # extras_require={
    #     'benchmark': benchmark_requires,
    #     'test': test_requires,
    #     'dev': dev_requires,
    # },
    packages=find_packages(),
    include_package_data=True,
)